from .checks.base import BaseCheck
from .checks.base import CheckLevel
from .config import Config
from .schema import ModelSchema
from .threedi_database import ThreediDatabase
from pathlib import Path
from typing import Iterator
from typing import NamedTuple
from typing import Optional
from typing import Set
from typing import Tuple


__all__ = ["ThreediModelChecker"]


class Context:
    def __init__(
        self,
        available_rasters: Optional[Set] = None,
        base_path: Optional[Path] = None,
    ):
        self.available_rasters = available_rasters
        self.base_path = base_path


class ThreediModelChecker:
    def __init__(self, threedi_db: ThreediDatabase, context: Optional[Context] = None):
        """Initialize the model checker.

        Optionally, supply the context of the model check:
        - "available_rasters": a set of raster options that are available
        - "base_path": file presence is checked relative to this path
        """
        self.db = threedi_db
        self.schema = ModelSchema(self.db)
        self.schema.validate_schema()
        self.config = Config(self.models)
        self.context = Context(**(context or {}))
        if not self.context.base_path:
            self.context.base_path = self.db.base_path

    @property
    def models(self):
        """Returns a list of declared models"""
        return self.schema.declared_models

    def errors(self, level=CheckLevel.ERROR) -> Iterator[Tuple[BaseCheck, NamedTuple]]:
        """Iterates and applies checks, returning any failing rows.

        By default, checks of WARNING and INFO level are ignored.

        :return: Tuple of the applied check and the failing row.
        """
        session = self.db.get_session()
        session.model_checker_context = self.context
        for check in self.checks(level=level):
            model_errors = check.get_invalid(session)
            for error_row in model_errors:
                yield check, error_row

    def checks(self, level=CheckLevel.ERROR) -> Iterator[BaseCheck]:
        """Iterates over all configured checks

        :return: implementations of BaseChecks
        """
        for check in self.config.iter_checks(level=level):
            yield check

    def check_table(self, table):
        pass

    def check_column(self, column):
        pass



from contextlib import contextmanager
from pathlib import Path
from sqlalchemy import create_engine
from sqlalchemy import event
from sqlalchemy.engine import Engine
from sqlalchemy.event import listen
from sqlalchemy.orm import sessionmaker

import shutil
import sqlite3
import tempfile


__all__ = ["ThreediDatabase"]


@event.listens_for(Engine, "connect")
def set_sqlite_pragma(dbapi_connection, connection_record):
    """Switch on legacy_alter_table setting to fix our migrations.

    Why?
    1) SQLite does not support "DROP COLUMN ...". You have to create a new table,
       copy the data, drop the old table, then rename. Luckily Alembic supports this pattern.
       They call it a "batch operation". See https://alembic.sqlalchemy.org/en/latest/batch.html.
    2) Newer SQLite drivers do a lot of fancy checks on a RENAME command. This made our
       "batch operations" fail in case a view referred to the table that is getting a "batch operation".
       The solution was a PRAGMA command. See https://www.sqlite.org/pragma.html#pragma_legacy_alter_table.
    """
    if isinstance(dbapi_connection, sqlite3.Connection):
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA legacy_alter_table=ON")
        # Some additional pragmas recommended in https://www.sqlite.org/security.html, paragraph 1.2
        cursor.execute("PRAGMA cell_size_check=ON")
        cursor.execute("PRAGMA mmap_size=0")
        cursor.close()


def load_spatialite(con, connection_record):
    """Load spatialite extension as described in
    https://geoalchemy-2.readthedocs.io/en/latest/spatialite_tutorial.html"""
    import sqlite3

    con.enable_load_extension(True)
    cur = con.cursor()
    libs = [
        # SpatiaLite >= 4.2 and Sqlite >= 3.7.17, should work on all platforms
        ("mod_spatialite", "sqlite3_modspatialite_init"),
        # SpatiaLite >= 4.2 and Sqlite < 3.7.17 (Travis)
        ("mod_spatialite.so", "sqlite3_modspatialite_init"),
        # SpatiaLite < 4.2 (linux)
        ("libspatialite.so", "sqlite3_extension_init"),
    ]
    found = False
    for lib, entry_point in libs:
        try:
            cur.execute("select load_extension('{}', '{}')".format(lib, entry_point))
        except sqlite3.OperationalError:
            continue
        else:
            found = True
            break
    if not found:
        raise RuntimeError("Cannot find any suitable spatialite module")
    cur.close()
    con.enable_load_extension(False)


class ThreediDatabase:
    def __init__(self, connection_settings, db_type="spatialite", echo=False):
        """

        :param connection_settings:
        db_type (str choice): database type. 'sqlite' and 'postgresql' are
                              supported
        """
        self.settings = connection_settings
        self.db_type = db_type
        self.echo = echo

        self._engine = None
        self._base_metadata = None

    @property
    def engine(self):
        return self.get_engine()

    @property
    def base_path(self):
        if self.db_type == "spatialite":
            return Path(self.settings["db_path"]).absolute().parent

    def get_engine(self, get_seperate_engine=False):

        if self._engine is None or get_seperate_engine:
            if self.db_type == "spatialite":
                engine = create_engine(
                    "sqlite:///{0}".format(self.settings["db_path"]), echo=self.echo
                )
                listen(engine, "connect", load_spatialite)
                if get_seperate_engine:
                    return engine
                else:
                    self._engine = engine

            elif self.db_type == "postgres":
                con = (
                    "postgresql://{username}:{password}@{host}:"
                    "{port}/{database}".format(**self.settings)
                )

                engine = create_engine(con, echo=self.echo)
                if get_seperate_engine:
                    return engine
                else:
                    self._engine = engine

        return self._engine

    def get_session(self, **kwargs):
        """Get a SQLAlchemy session for optimal control.

        It is probably necessary to call ``session.commit``, ``session.rollback``
        and/or ``session.close``.

        See also:
          https://docs.sqlalchemy.org/en/13/orm/session_basics.html
        """
        return sessionmaker(bind=self.engine)(**kwargs)

    @contextmanager
    def session_scope(self, **kwargs):
        """Get a session to execute a single transaction in a "with as" block."""
        session = self.get_session(**kwargs)
        try:
            yield session
            session.commit()
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

    @contextmanager
    def file_transaction(self, **kwargs):
        """Copy the complete database into a tmpdir and work on that one.

        On contextmanager exit, the database is copied back and the real
        database is overwritten. On error, nothing happens.
        """
        if self.db_type != "spatialite":
            raise NotImplementedError(
                f"Cannot make database backups for db_type {self.db_type}"
            )
        with tempfile.NamedTemporaryFile(suffix=".sqlite") as work_file:
            # copy the database to the temporary directory
            shutil.copy(self.settings["db_path"], work_file.name)
            # yield a new ThreediDatabase refering to the backup
            yield self.__class__({"db_path": work_file.name}, "spatialite")
            # this is only reached if there was no error:
            shutil.copy(work_file.name, self.settings["db_path"])

    def check_connection(self):
        """Check if there a connection can be started with the database

        :return: True if a connection can be established, otherwise raises an
            appropriate error.
        """
        session = self.get_session()
        r = session.execute("select 1")
        return r.fetchone()

    def check_integrity(self):
        """Should be called before doing anything with an untrusted sqlite file."""
        with self.session_scope() as session:
            session.execute("PRAGMA integrity_check")



from threedi_modelchecker import exporters
from threedi_modelchecker.checks.base import CheckLevel
from threedi_modelchecker.model_checks import ThreediModelChecker
from threedi_modelchecker.schema import ModelSchema
from threedi_modelchecker.threedi_database import ThreediDatabase

import click


@click.group()
@click.option(
    "-s",
    "--sqlite",
    type=click.Path(exists=True, readable=True),
    help="Path to an sqlite (spatialite) file",
)
@click.option("-d", "--database", help="PostGIS database name to connect to")
@click.option("-h", "--host", help="PostGIS database server host")
@click.option("-p", "--port", default=5432, help="PostGIS database server port")
@click.option("-u", "--username", help="PostGIS database username")
@click.option("-u", "--password", help="PostGIS database password")
@click.pass_context
def threedi_modelchecker(ctx, sqlite, database, host, port, username, password):
    """Checks the threedi-model for errors / warnings / info messages"""
    ctx.ensure_object(dict)

    if sqlite:
        sqlite_settings = {"db_path": sqlite, "db_file": sqlite}
        db = ThreediDatabase(
            connection_settings=sqlite_settings, db_type="spatialite", echo=False
        )
    else:
        postgis_settings = {
            "host": host,
            "port": port,
            "database": database,
            "username": username,
            "password": password,
        }
        db = ThreediDatabase(
            connection_settings=postgis_settings, db_type="postgres", echo=False
        )
    ctx.obj["db"] = db


@threedi_modelchecker.command()
@click.option("-f", "--file", help="Write errors to file, instead of stdout")
@click.option(
    "-l",
    "--level",
    type=click.Choice([x.name for x in CheckLevel], case_sensitive=False),
    default="ERROR",
    help="Minimum check level.",
)
@click.pass_context
def check(ctx, file, level):
    """Checks the threedi model schematisation for errors."""
    level = level.upper()
    if level == "ERROR":
        msg = "errors"
    elif level == "WARNING":
        msg = "errors or warnings"
    else:
        msg = "errors, warnings or info messages"
    click.echo("Parsing threedi-model for any %s" % msg)
    if file:
        click.echo("Model errors will be written to %s" % file)

    mc = ThreediModelChecker(ctx.obj["db"])
    model_errors = mc.errors(level=level)

    if file:
        exporters.export_to_file(model_errors, file)
    else:
        exporters.print_errors(model_errors)

    click.echo("Finished processing model")


@threedi_modelchecker.command()
@click.option(
    "-r", "--revision", default="head", help="The schema revision to migrate to"
)
@click.pass_context
def migrate(ctx, revision):
    """Migrate the threedi model schematisation to the latest version."""
    schema = ModelSchema(ctx.obj["db"])
    click.echo("The current schema revision is: %s" % schema.get_version())
    click.echo("Running alembic upgrade script...")
    schema.upgrade(revision=revision)
    click.echo("The migrated schema revision is: %s" % schema.get_version())


if __name__ == "__main__":
    threedi_modelchecker()



from sqlalchemy.types import Integer
from sqlalchemy.types import TypeDecorator
from sqlalchemy.types import VARCHAR


class CustomEnum(TypeDecorator):
    """Column type for storing integer python enums.

    The CustomEnum will make use of the backend 'impl' datatype to store
    the data. Data is converted back it's enum_class python object if possible.
    If not, the raw 'impl' is returned.

    CustomEnum should not be used directly but inherited by a class which
    provides the `impl` datatype.
    """

    def __init__(self, enum_class):
        """

        :param enum_class: instance of enum.Enum
        """
        super().__init__()
        self.enum_class = enum_class
        self.enums = [e.value for e in enum_class]

    def process_bind_param(self, value, dialect):
        if isinstance(value, self.enum_class):
            return value.value
        else:
            return value

    def process_result_value(self, value, dialect):
        if value in self.enums:
            return self.enum_class(value)
        else:
            return value


class IntegerEnum(CustomEnum):
    """Column type for storing integer python enums.

    The CustomEnum will make use of the backend 'impl' datatype to store
    the data. Data is converted back it's enum_class python object if possible.
    If not, the raw 'impl' is returned.

    CustomEnum should not be used directly but inherited by a class which
    provides the `impl` datatype.
    """

    cache_ok = True
    impl = Integer


class VarcharEnum(CustomEnum):
    cache_ok = True
    impl = VARCHAR



from .errors import MigrationMissingError
from .threedi_model import constants
from .threedi_model import models
from alembic.config import Config
from alembic.environment import EnvironmentContext
from alembic.migration import MigrationContext
from alembic.script import ScriptDirectory
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import Table

import warnings


__all__ = ["ModelSchema"]


def get_alembic_config(connection=None):
    alembic_cfg = Config()
    alembic_cfg.set_main_option("script_location", "threedi_modelchecker:migrations")
    alembic_cfg.set_main_option("version_table", constants.VERSION_TABLE_NAME)
    if connection is not None:
        alembic_cfg.attributes["connection"] = connection
    return alembic_cfg


def get_schema_version():
    """Returns the version of the schema in this library"""
    config = get_alembic_config()
    script = ScriptDirectory.from_config(config)
    with EnvironmentContext(config=config, script=script) as env:
        return int(env.get_head_revision())


def _upgrade_database(db, revision="head"):
    """Upgrade ThreediDatabase instance"""
    with db.get_engine().begin() as connection:
        config = get_alembic_config(connection)
        script = ScriptDirectory.from_config(config)

        def upgrade(rev, context):
            return script._upgrade_revs(revision, rev)

        with EnvironmentContext(
            config,
            script,
            fn=upgrade,
            destination_rev=revision,
            version_table=constants.VERSION_TABLE_NAME,
        ):
            script.run_env()


class ModelSchema:
    def __init__(self, threedi_db, declared_models=models.DECLARED_MODELS):
        self.db = threedi_db
        self.declared_models = declared_models

    def _get_version_old(self):
        """The version of the database using the old 'south' versioning."""
        south_migrationhistory = Table(
            "south_migrationhistory", MetaData(), Column("id", Integer)
        )
        engine = self.db.get_engine()
        if not engine.has_table("south_migrationhistory"):
            return
        with engine.connect() as connection:
            query = south_migrationhistory.select().order_by(
                south_migrationhistory.columns["id"].desc()
            )
            versions = list(connection.execute(query.limit(1)))
            if len(versions) == 1:
                return versions[0][0]
            else:
                return None

    def get_version(self):
        """Returns the id (integer) of the latest migration"""
        with self.db.get_engine().connect() as connection:
            context = MigrationContext.configure(
                connection, opts={"version_table": constants.VERSION_TABLE_NAME}
            )
            version = context.get_current_revision()

        if version is not None:
            return int(version)
        else:
            return self._get_version_old()

    def upgrade(self, revision="head", backup=True):
        """Upgrade the database to the latest version.

        This requires either a completely empty database or a database with its
        current schema version at least 174 (the latest migration of the old
        model databank).

        The upgrade is done using database transactions. However, for SQLite,
        database transactions are only partially supported. To ensure that the
        database file does not become corrupt, enable the "backup" parameter.
        If the database is temporary already (or if it is PostGIS), disable
        it.
        """
        v = self.get_version()
        if v is not None and v < constants.LATEST_SOUTH_MIGRATION_ID:
            raise MigrationMissingError(
                f"The modelchecker cannot update versions below "
                f"{constants.LATEST_SOUTH_MIGRATION_ID}. Please consult the "
                f"3Di documentation on how to update legacy databases."
            )
        if backup:
            with self.db.file_transaction() as work_db:
                _upgrade_database(work_db, revision=revision)
        else:
            _upgrade_database(self.db, revision=revision)

    def validate_schema(self):
        """Very basic validation of 3Di schema.

        Check that the database has the latest migration applied. If the
        latest migrations is applied, we assume the database also contains all
        tables and columns defined in threedi_model.models.py.

        :return: True if the threedi_db schema is valid, raises an error otherwise.
        :raise MigrationMissingError, MigrationTooHighError
        """
        version = self.get_version()
        if version is None or version < constants.MIN_SCHEMA_VERSION:
            raise MigrationMissingError(
                f"The modelchecker requires at least schema version "
                f"{constants.MIN_SCHEMA_VERSION}. Current version: {version}."
            )

        schema_version = get_schema_version()
        if version > schema_version:
            warnings.warn(
                f"The database version is higher than the modelchecker "
                f"({version} > {schema_version}). This may lead to unexpected "
                f"results. "
            )
        return True

    def get_missing_tables(self):
        pass

    def get_missing_columns(self):
        pass
